// liaison.cc 

#include "copyright.h"
#include "liaison.h"
#include "system.h"
#include <unistd.h>

#include <strings.h>

#define LARGEUR_FENETRE_EMISSION  5
#define LARGEUR_FENETRE_RECEPTION 2

// Les 3 fonctions suivantes servent juste  rediriger l'excution des
// Threads correspondants vers les bonnes fonctions C++...

static void
ProtocoleReception_Root (int arg)
{
  ((CoucheLiaison *) arg)->ProtocoleReception ();
}

static void
ProtocoleEmission_Root (int arg)
{
  ((CoucheLiaison *) arg)->ProtocoleEmission ();
}

static void
DemonReception_Root (int arg)
{
  ((CoucheLiaison *) arg)->DemonReception ();
}


CoucheLiaison::CoucheLiaison (NetworkAddress addr, double reliability)
{
  Thread *tr, *ts, *tf;

  couchePhysique = new CouchePhysique (addr, reliability);

  tamponReception = new Tampon (LARGEUR_FENETRE_RECEPTION);
  tamponEmission = new Tampon (LARGEUR_FENETRE_EMISSION);
  tamponDemon = new SynchList ();

  temporisateur = new Temporisateur;

  tr = new Thread ("protocole de reception");
  tr->Fork (ProtocoleReception_Root, (int) this);

  ts = new Thread ("protocole d'mission");
  ts->Fork (ProtocoleEmission_Root, (int) this);

  tf = new Thread ("demon de rception");
  tf->Fork (DemonReception_Root, (int) this);
}

CoucheLiaison::~CoucheLiaison ()
{
  delete couchePhysique;
  delete tamponReception;
  delete tamponEmission;
  delete tamponDemon;
  delete temporisateur;
}

void
CoucheLiaison::EnvoyerPaquet (char *data, unsigned len)
{
  Trame *trame = new Trame (Trame_INFO, data, len);

  tamponEmission->Ajouter (trame);
  evtEmission.Generer (new Evenement (Event_NEW_FRAME));
}

void
CoucheLiaison::RecevoirPaquet (char *data, unsigned *len)
{
  Trame *trame;

  trame = tamponReception->Retirer ();

  bcopy (trame->data, data, trame->header.len);
  *len = trame->header.len;

  delete trame;
}

void
CoucheLiaison::Emettre (unsigned trame_courante)
{
  Trame *trame;

  // On rcupre la trame du tampon d'mission
  trame = tamponEmission->TrameNumero (trame_courante);

  trame->header.NS = trame_courante;
  trame->header.DebugInfo = trame_courante; // inutile au protocole

  // On l'envoie sur le rseau
  DEBUG ('l', "\t [E] Emission de la trame %d\n", trame_courante);
  couchePhysique->EnvoyerTrame (trame);

  temporisateur->Armer (2 * TEMPO_ACQUITTEMENT, Expiration_ACK,
			trame_courante, &evtEmission);
}

void
CoucheLiaison::ProtocoleEmission ()
{

  Evenement *evt;
  unsigned ack_attendu = 0;  // numros de squence  usage local
  unsigned trame_courante = 0;	


  for (;;)
    {

      evt = evtEmission.Recuperer ();	// On attend le prochain
                                        // vnement
      switch (evt->type)
	{

	case Event_NEW_FRAME:
	  {
	    // La couche suprieure a russi a dposer une trame
	    // supplmentaire dans le tampon de
	    // LARGEUR_FENETRE_EMISSION elt => on peut l'envoyer
	    
	    Emettre (trame_courante);
	    trame_courante++;
	    
	    break;
	  }
      
	case Event_ACK_RECEIVED:
	  {

	    Trame *trame_ack = (Trame *) tamponDemon->Remove ();

	    if (trame_ack->header.NR == ack_attendu)
	      {

		// On reoit un acquittement
		DEBUG ('l', "\t [E] Acquittement reu pour la trame %d\n",
		       ack_attendu);

		ack_attendu++;
		
		// On peut retirer la trame mise du tampon et la
		// dtruire...

		delete tamponEmission->Retirer ();
	      }
	    
	    else
	      {
		// C'est un acquittement reu "hors-squence"
		DEBUG ('l', "\t [E] Acquittement %d ignor (attendu %d) \n",
		       trame_ack->header.NR, ack_attendu);
	      }
	    delete trame_ack;
	    break;
	  }
	  
	case Event_TIMER_EXPIRATION:
	  {
	    // expiration d'un temporisateur

	    if (evt->num >= ack_attendu)
	      {

		DEBUG ('l',
		       "\t [E] Expiration du temporisateur pour la trame %d\n",
		       evt->num);

		// Il faut r-mettre perdue...
		Emettre (evt->num);
	      }

	    break;
	  }

	default:
	  {
	    fprintf (stderr, "Oups: vnement inattendu en mission !\n");
	    interrupt->Halt ();
	  }
	}			// switch

      delete evt;
    }
}

void
CoucheLiaison::ProtocoleReception ()
{
  Evenement *evt;
  unsigned numero_attendu = 0;

  for (;;)
    {

      // On attend le prochain vnement
      evt = evtReception.Recuperer ();

      switch (evt->type)
	{

	case Event_NEW_FRAME:
	  {

	    Trame *trame_ack = new Trame (Trame_ACK);
	    Trame *trame_recue = (Trame *) tamponDemon->Remove ();

	    trame_ack->header.NR = trame_recue->header.NS;
	    trame_ack->header.DebugInfo = trame_recue->header.DebugInfo;

	    if (trame_recue->header.NS == numero_attendu)
	      {
		if(tamponReception->TenterAjout (trame_recue) == -1)
		  {
		    // Oups! Le tampon est plein...
		    DEBUG ('l', "\t [R] Trame %d dtruite (tampon satur)\n",
			   trame_recue->header.NS);
		    delete trame_recue;

		    break; // pour ne pas envoyer d'ack
		  }
		numero_attendu++;
	      }
	    else
	      {
		DEBUG ('l', "\t [R] Trame %d ignore (attendue %d)\n",
		       trame_recue->header.NS, numero_attendu);
		delete trame_recue;
	      }

	    DEBUG ('l', "\t [R] Emission d'un ack %d (trame %d)\n",
		   trame_ack->header.NR, trame_ack->header.DebugInfo);
	    couchePhysique->EnvoyerTrame (trame_ack);
	    break;
	  }

	default:
	  {
	    fprintf (stderr, "Oups: vnement inattendu en rception !\n");
	    interrupt->Halt ();
	  }

	}

      delete evt;
    }
}

void
CoucheLiaison::DemonReception ()
{

  for (;;)
    {

      Trame *trame = couchePhysique->RecevoirTrame ();
      
      tamponDemon->Append (trame);

      switch (trame->header.type)
	{
	case Trame_INFO:
	  {

	    DEBUG ('l', "\t [D] Rception de la trame %d \n",
		   trame->header.DebugInfo);

	    evtReception.
	      Generer (new Evenement (Event_NEW_FRAME));

	    break;
	  }
	case Trame_ACK:
	  {
	    DEBUG ('l', "\t [D] Rception de l'ack %d (trame %d) \n",
		   trame->header.NR, trame->header.DebugInfo);

	    // On signale l'vnement au dmon de rception
	    evtEmission.
	      Generer (new Evenement (Event_ACK_RECEIVED));
	    break;
	  }

	case Trame_REJECT:
	  {
	    ASSERT (FALSE);
	    break;
	  }

	default:
	  {
	    fprintf (stderr, "Oups: trame de type inconnu reue !\n");
	    interrupt->Halt ();
	  }
	}			// switch
            
    }				// for
}
