// nettest.cc
//
//      Test de la couche "liaison" tablie entre deux machines Nachos
//
//      Il faut donc lancer deux processus Nachos quasi-simultanment,
//      de prfrence dans deux fentres diffrentes.
//
//      # Dans la premire fentre :
//      ./nachos -m 0
//
//      # Dans la seconde fentre :
//      ./nachos -m 1
//
//      Pour introduire des pertes de paquets, il suffit d'utiliser
//      l'option '-l <taux>' qui permet de spcifier le taux de
//      fiabilit de la ligne _en_mission_. Voici par exemple comment
//      mettre en place une liaison perdant environ 50% de ses paquets
//      de m0 vers m1 et 10% dans l'autre sens :
//
//      ./nachos -m 0 -l 0.5
//
//      ./nachos -m 1 -l 0.9
//
//      Notez enfin que l'option '-d n' permet l'affichage des
//      messages de "debug" relatifs au rseau...
//

#include "copyright.h"

#include "system.h"
#include "liaison.h"

#define NB_PAQUETS  10

static void sleepHandler(int arg) {
  scheduler->ReadyToRun((Thread*)arg);
}

static void do_UserSleep(int time) {
  IntStatus oldLevel = interrupt->SetLevel(IntOff);
  interrupt->Schedule(&sleepHandler, (int) currentThread,
                      time, TimerInt);
  currentThread->Sleep();
  (void) interrupt->SetLevel(oldLevel);
}

void
NetTest (void)
{
  char data[1024] = "Hello there from!";
  char buffer[MaxPacketSize];
  unsigned i, len, self;

  printf ("*** Dbut du main\n");

  self = coucheLiaison->GetNetAddress ();

  if (self == 0)
    {				// machine 0

      for (i = 0; i < NB_PAQUETS; i++)
	{

	  sprintf (data, "Hello from %d (paquet %d)", self, i);

	  // Emission
	  printf ("J'ai tent d'mettre le paquet <%s>\n", data);
	  coucheLiaison->EnvoyerPaquet (data, strlen (data) + 1);
	}

    }
  else
    {				// machine 1

      for (i = 0; /* i < NB_PAQUETS */ ; i++)
	{

	  // Reception
	  coucheLiaison->RecevoirPaquet (buffer, &len);
	  printf ("J'ai reu le paquet <%s>\n", buffer);
	  do_UserSleep(1000000);//on simule une priode de calcul.
	}

    }

  printf ("*** Fin du main\n");
}
